/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2009 Alpha Data
**
** dma2.c - IOCTL handlers for the DMA API functions
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "drvparam.h"
#include "ioctls.h"
#include "dma.h"
#include "dmadesc.h"
#include "dmacore.h"

/*
** This callback function is attached to the DMANODEs created by
** the DoDMA and DoDMAImmediate IOCTL handlers.
**
** May be called in an arbitrary context.
*/
static void
doDMACallback(
  DeviceContext*  devCtx,
  DmaNode* pNode)
{
  OS_EVENT* pEvent;

  pEvent = (OS_EVENT*)pNode->client.pContext;
  os_signal_event(pEvent);
}

/*
** Handle ADMXRC_IOCTL_DODMA
**
** Must be called in a thread context.
*/
OS_STATUS
ADMXRC_DoDMA(
  DeviceContext* devCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_DODMA* pIoctl,
  unsigned long inSize,
  unsigned long outSize)
{
  DmaNode* pDmaNode = NULL;
  DmaTarget* pDmaTgt = NULL;
  OS_STATUS status = OS_STATUS_SUCCESS;
  os_uint32 local;
  unsigned long length;
  os_uint32 mode;
  unsigned long flags;
  unsigned long channel;
  unsigned long dir;
  unsigned long offset, limit;
  unsigned char queue;
  OS_EVENT event;
  os_boolean bMustStart;
  os_boolean bUseTimeout;
  os_ticks start, now, expire, ticks; /* ticks */
  unsigned long remaining_us; /* microseconds */
  os_boolean bRemote;
  os_boolean bSignalled, bCancelled;

  if (pIoctl == NULL || inSize != sizeof(pIoctl->in) || outSize != sizeof(pIoctl->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  if (!devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  }

  /* Validate PLX DMA mode */
  mode = pIoctl->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = pIoctl->in.direction;
  if (dir != ADMXRC2_PCITOLOCAL && dir != ADMXRC2_LOCALTOPCI) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  local = pIoctl->in.local;
  length = pIoctl->in.length;
  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate flags parameter */
  flags = pIoctl->in.flags;
  if (flags & ~ADMXRC2_DMAFLAG_DONOTQUEUE) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate channel parameter */
  channel = pIoctl->in.channel;
  if (channel != ADMXRC2_DMACHAN_ANY && channel >= devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  pDmaNode = (DmaNode*)os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (pDmaNode == NULL) {
    return OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
  }

  /* Get values indicating whether or not to use a timeout */
  bUseTimeout = pIoctl->in.bUseTimeout;
  if (bUseTimeout) {
    start = os_get_time();
    expire = start + os_us_to_ticks(pIoctl->in.timeout);
  } else {
    start = expire = 0; /* Stop compiler complaining about uninit vars. */
  }

  pDmaTgt = ADMXRC_GetDMATarget(devCtx, pIoctl->in.dmadesc);
  if (pDmaTgt == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_DMADESC);
    goto cleanup;
  }
  bRemote = pDmaTgt->bRemote;

  /* Validate the offset and length parameters */
  offset = pIoctl->in.offset;
  limit = offset + length;
  if (limit < offset || limit > pDmaTgt->length) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
    goto cleanup;
  }

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC2_DMACHAN_ANY) ? devCtx->info.cardInfo.numDmaChan : (unsigned char)channel;

  /* Prepare the node for insertion into DMA queue */
  pDmaNode->client.pContext = (void*)&event;
  pDmaNode->client.pCallback = doDMACallback;
  pDmaNode->client.pBuffer = pDmaTgt;
  pDmaNode->client.mode = mode;
  pDmaNode->client.local = local;
  pDmaNode->client.length = length;
  pDmaNode->client.offset = offset;
  pDmaNode->client.direction = (unsigned char)dir;
  pDmaNode->client.queue = queue;
  pDmaNode->client.bTwoD = FALSE;

  os_dbgprint(6, ("ADMXRC_DoDMA: enqueuing DMANODE 0x%p dmadesc=%lu offset=%lu length=%lu local=0x%lx dir=%lu channel=%lu mode=0x%lx flags=0x%lx queue=%lu\n",
		  (void*)pDmaNode,
		  (unsigned long)pIoctl->in.dmadesc,
		  (unsigned long)offset,
		  (unsigned long)length,
		  (unsigned long)local,
		  (unsigned long)dir,
		  (unsigned long)channel,
		  (unsigned long)mode,
		  (unsigned long)flags,
		  (unsigned long)queue));

  status = OS_STATUS_SUCCESS;
  pReq->user.context = (void*)pDmaNode;
  os_init_event(&event);

  bMustStart = (flags & ADMXRC2_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, pDmaNode, bMustStart)) {
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
    goto completed;
  }

  if (bUseTimeout) {
    now = os_get_time();
    if (os_time_after(now, expire)) {
      ticks = 0;
    } else {
      ticks = expire - now;
    }
    bCancelled = os_wait_event_timeout(&event, ticks, TRUE, &bSignalled);
    if (bCancelled) {
      if (bSignalled) {
	status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      } else {
	status = OS_MAKE_STATUS(ADMXRC2_TIMEOUT);
      }
      if (!ADMXRC_CancelDma(devCtx, pDmaNode)) {
	os_wait_event(&event, FALSE);
      }
      goto completed;
    }
  } else {
    bSignalled = os_wait_event(&event, TRUE);
    if (bSignalled) {
      status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      if (!ADMXRC_CancelDma(devCtx, pDmaNode)) {
	os_wait_event(&event, FALSE);
      }
      goto completed;
    }
  }
  
 completed:
  if (OS_SUCCESS(status)) {
    pReq->status.length = sizeof(pIoctl->out);
    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after(now, expire)) {
	pIoctl->out.remaining = 0;
      } else {
	remaining_us = os_ticks_to_us(expire - now);
	if (remaining_us > pIoctl->in.timeout) {
	  pIoctl->out.remaining = pIoctl->in.timeout;
	} else {
	  pIoctl->out.remaining = remaining_us;
	}
      }
    }
  }

 cleanup:
  if (pDmaTgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, pDmaTgt);
  }
  if (pDmaNode != NULL) {
    os_free_lcache(devCtx->dma.ctlPool, pDmaNode);
  }

  return status;
}

/*
** Handle ADMXRC_IOCTL_DODMA2D
**
** Must be called in a thread context.
*/
OS_STATUS
ADMXRC_DoDMA2D(
  DeviceContext* devCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_DODMA2D* pIoctl,
  unsigned long inSize,
  unsigned long outSize)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  DmaNode* pDmaNode = NULL;
  DmaTarget* pDmaTgt = NULL;
  unsigned long spanned, local_spanned;
  os_uint32 local;
  unsigned long width;
  unsigned long skip, localSkip;
  unsigned long length;
  os_uint32 mode;
  unsigned long flags;
  unsigned long channel;
  unsigned long dir;
  unsigned long offset, limit;
  unsigned char queue;
  OS_EVENT event;
  os_boolean bMustStart;
  os_boolean bUseTimeout, bCancelled, bSignalled;
  os_ticks start, now, expire, ticks; /* ticks */
  unsigned long remaining_us; /* microseconds */

  if (pIoctl == NULL || inSize != sizeof(pIoctl->in) || outSize != sizeof(pIoctl->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  if (!devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  }

  /* Validate PLX DMA mode */
  mode = pIoctl->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = pIoctl->in.direction;
  if (dir != ADMXRC2_PCITOLOCAL && dir != ADMXRC2_LOCALTOPCI) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  local = pIoctl->in.local;
  length = pIoctl->in.length;
  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate the 2D parameters */
  {
    unsigned long n_full_strip, n_skip;
    unsigned long partial_strip;

    width = pIoctl->in.width;
    skip = pIoctl->in.skip;
    localSkip = pIoctl->in.local_skip;

    n_full_strip = length / width;
    partial_strip = length % width;
    if (partial_strip) {
      n_skip = n_full_strip;
    } else {
      n_skip = n_full_strip - 1;
    }
    spanned = width * n_full_strip + partial_strip;
    local_spanned = spanned + n_skip * localSkip;
    spanned += n_skip * skip;
  }

  /* Validate flags parameter */
  flags = pIoctl->in.flags;
  if (flags & ~ADMXRC2_DMAFLAG_DONOTQUEUE) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate channel parameter */
  channel = pIoctl->in.channel;
  if (channel != ADMXRC2_DMACHAN_ANY && channel >= devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  pDmaNode = (DmaNode*) os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (pDmaNode == NULL) {
    return OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
  }

  /* Get values indicating whether or not to use a timeout */
  bUseTimeout = pIoctl->in.bUseTimeout;
  if (bUseTimeout) {
    start = os_get_time();
    expire = start + os_us_to_ticks(pIoctl->in.timeout);
  } else {
    start = expire = 0; /* Stop compiler complaining about uninit vars. */
  }

  /* Lock down the user buffer and obtain an MDL for it */
  pDmaTgt = ADMXRC_GetDMATarget(devCtx, pIoctl->in.dmadesc);
  if (pDmaTgt == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_DMADESC);
    goto cleanup;
  }
  if (pDmaTgt->bRemote) {
    status = OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
    goto cleanup;
  }

  /* Validate the offset and length parameters */
  offset = pIoctl->in.offset;
  limit = offset + spanned;
  if (limit < offset || offset > pDmaTgt->length || limit > pDmaTgt->length) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
    goto cleanup;
  }

  /* Determine which DMA queue to add this node to */
  queue = (channel == ADMXRC2_DMACHAN_ANY) ? devCtx->info.cardInfo.numDmaChan : (unsigned char)channel;

  /* Prepare the node for insertion into DMA queue */
  pDmaNode->client.pContext = (void*)&event;
  pDmaNode->client.pCallback = doDMACallback;
  pDmaNode->client.pBuffer = pDmaTgt;
  pDmaNode->client.mode = mode;
  pDmaNode->client.local = local;
  pDmaNode->client.length = length;
  pDmaNode->client.offset = offset;
  pDmaNode->client.variant.twoD.width = width;
  pDmaNode->client.variant.twoD.skip = skip;
  pDmaNode->client.variant.twoD.localSkip = localSkip;
  pDmaNode->client.direction = (unsigned char)dir;
  pDmaNode->client.queue = queue;
  pDmaNode->client.bTwoD = TRUE;

  os_dbgprint(6, ("ADMXRC_DoDMA2D: enqueuing DMANODE 0x%p dmadesc=%ld offset=%ld length=%ld local=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*)pDmaNode,
		  (unsigned long)pIoctl->in.dmadesc,
		  (unsigned long)offset,
		  (unsigned long)length,
		  (unsigned long)local,
		  (unsigned long)dir,
		  (unsigned long)channel,
		  (unsigned long)mode,
		  (unsigned long)flags,
		  (unsigned long)queue));

  status = OS_STATUS_SUCCESS;
  pReq->user.context = (void*) pDmaNode;
  os_init_event(&event);

  bMustStart = (flags & ADMXRC2_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, pDmaNode, bMustStart)) {
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
    goto completed;
  }

  if (bUseTimeout) {
    now = os_get_time();
    if (os_time_after(now, expire)) {
      ticks = 0;
    } else {
      ticks = expire - now;
    }
    bCancelled = os_wait_event_timeout(&event, ticks, TRUE, &bSignalled);
    if (bCancelled) {
      if (bSignalled) {
	status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      } else {
	status = OS_MAKE_STATUS(ADMXRC2_TIMEOUT);
      }
      if (!ADMXRC_CancelDma(devCtx, pDmaNode)) {
	os_wait_event(&event, FALSE);
      }
      goto completed;
    }
  } else {
    bSignalled = os_wait_event(&event, TRUE);
    if (bSignalled) {
      status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      if (!ADMXRC_CancelDma(devCtx, pDmaNode)) {
	os_wait_event(&event, FALSE);
      }
      goto completed;
    }
  }

 completed:
  if (OS_SUCCESS(status)) {
    pReq->status.length = sizeof(pIoctl->out);
    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after(now, expire)) {
	pIoctl->out.remaining = 0;
      } else {
	remaining_us = os_ticks_to_us(expire - now);
	if (remaining_us > pIoctl->in.timeout) {
	  pIoctl->out.remaining = pIoctl->in.timeout;
	} else {
	  pIoctl->out.remaining = remaining_us;
	}
      }
    }
  }

 cleanup:
  if (pDmaTgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, pDmaTgt);
  }
  if (pDmaNode != NULL) {
    os_free_lcache(devCtx->dma.ctlPool, pDmaNode);
  }

  return status;
}

/*
** Handle ADMXRC_IOCTL_DODMAIMMEDIATE
**
** Must be called in a thread context.
*/
OS_STATUS
ADMXRC_DoDMAImmediate(
  DeviceContext* devCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_DODMAIMMEDIATE* pIoctl,
  unsigned long inSize,
  unsigned long outSize)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  DmaNode* pDmaNode;
  DmaTarget* pDmaTgt = NULL;
  os_uint32 local;
  unsigned long length;
  os_uint32 mode;
  unsigned long flags;
  unsigned long channel;
  unsigned long dir;
  unsigned char queue;
  void* pBuffer;
  OS_EVENT event;
  os_boolean bMustStart;
  os_boolean bUseTimeout, bCancelled, bSignalled;
  os_ticks start, now, expire, ticks; /* ticks */
  unsigned long remaining_us; /* microseconds */

  if (pIoctl == NULL || inSize != sizeof(pIoctl->in) || outSize != sizeof(pIoctl->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  if (!devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  }

  /* Validate PLX DMA mode */
  mode = pIoctl->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = pIoctl->in.direction;
  if (dir != ADMXRC2_PCITOLOCAL && dir != ADMXRC2_LOCALTOPCI) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  pBuffer = pIoctl->in.buffer;
  local = pIoctl->in.local;
  length = pIoctl->in.length;
  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate flags parameter */
  flags = pIoctl->in.flags;
  if (flags & ~ADMXRC2_DMAFLAG_DONOTQUEUE) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate channel parameter */
  channel = pIoctl->in.channel;
  if (channel != ADMXRC2_DMACHAN_ANY && channel >= devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  pDmaNode = (DmaNode*) os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (pDmaNode == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
    goto cleanup;
  }

  /* Get values indicating whether or not to use a timeout */
  bUseTimeout = pIoctl->in.bUseTimeout;
  if (bUseTimeout) {
    start = os_get_time();
    expire = start + os_us_to_ticks(pIoctl->in.timeout);
  } else {
    start = expire = 0; /* Stop compiler complaining about uninit vars. */
  }

  /* Lock down the user buffer and obtain an MDL for it */
  status = ADMXRC_CreateDMATarget(devCtx, pBuffer, length, &pDmaTgt);
  if (!OS_SUCCESS(status)) {
    goto cleanup;
  }

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC2_DMACHAN_ANY) ? devCtx->info.cardInfo.numDmaChan : (unsigned char) channel;

  /* Prepare the node for insertion into DMA queue */
  pDmaNode->client.pContext = (void*) &event;
  pDmaNode->client.pCallback = doDMACallback;
  pDmaNode->client.pBuffer = pDmaTgt;
  pDmaNode->client.mode = mode;
  pDmaNode->client.local = local;
  pDmaNode->client.length = length;
  pDmaNode->client.offset = 0;
  pDmaNode->client.direction = (unsigned char)dir;
  pDmaNode->client.queue = queue;
  pDmaNode->client.bTwoD = FALSE;

  os_dbgprint(6, ("ADMXRC_DoDMAImmediate: pDmaNode=0x%p pBuffer=0x%p pDmaTgt=0x%p length=0x%lu local=0x%lx dir=%lu channel=%lu mode=0x%lx flags=0x%lx queue=%lu\n",
		  (void*)pDmaNode,
		  (void*)pBuffer,
		  (void*)pDmaTgt,
		  (unsigned long)length,
		  (unsigned long)local,
		  (unsigned long)dir,
		  (unsigned long)channel,
		  (unsigned long)mode,
		  (unsigned long)flags,
		  (unsigned long)queue));

  status = OS_STATUS_SUCCESS;
  pReq->user.context = (void*) pDmaNode;
  os_init_event(&event);

  bMustStart = (flags & ADMXRC2_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, pDmaNode, bMustStart)) {
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
    goto completed;
  }

  if (bUseTimeout) {
    now = os_get_time();
    if (os_time_after(now, expire)) {
      ticks = 0;
    } else {
      ticks = expire - now;
    }
    bCancelled = os_wait_event_timeout(&event, ticks, TRUE, &bSignalled);
    if (bCancelled) {
      if (bSignalled) {
	status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      } else {
	status = OS_MAKE_STATUS(ADMXRC2_TIMEOUT);
      }
      if (!ADMXRC_CancelDma(devCtx, pDmaNode)) {
	os_wait_event(&event, FALSE);
      }
      goto completed;
    }
  } else {
    bSignalled = os_wait_event(&event, TRUE);
    if (bSignalled) {
      status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      if (!ADMXRC_CancelDma(devCtx, pDmaNode)) {
	os_wait_event(&event, FALSE);
      }
      goto completed;
    }
  }
  
 completed:
  if (OS_SUCCESS(status)) {
    pReq->status.length = sizeof(pIoctl->out);
    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after(now, expire)) {
	pIoctl->out.remaining = 0;
      } else {
	remaining_us = os_ticks_to_us(expire - now);
	if (remaining_us > pIoctl->in.timeout) {
	  pIoctl->out.remaining = pIoctl->in.timeout;
	} else {
	  pIoctl->out.remaining = remaining_us;
	}
      }
    }
  }

 cleanup:
  if (pDmaTgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, pDmaTgt);
  }
  if (pDmaNode != NULL) {
    os_free_lcache(devCtx->dma.ctlPool, pDmaNode);
  }

  return status;
}

/*
** Handle ADMXRC_IOCTL_DODMAIMMEDIATE2D
**
** Must be called in a thread context.
*/
OS_STATUS
ADMXRC_DoDMAImmediate2D(
  DeviceContext* devCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_DODMAIMMEDIATE2D* pIoctl,
  unsigned long inSize,
  unsigned long outSize)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  DmaNode* pDmaNode;
  DmaTarget* pDmaTgt = NULL;
  unsigned long spanned, local_spanned;
  os_uint32 local;
  unsigned long width;
  unsigned long skip, localSkip;
  unsigned long length;
  os_uint32 mode;
  unsigned long flags;
  unsigned long channel;
  unsigned long dir;
  unsigned char queue;
  void* pBuffer;
  OS_EVENT event;
  os_boolean bMustStart;
  os_boolean bUseTimeout;
  os_ticks start, now, expire, ticks; /* ticks */
  unsigned long remaining_us; /* microseconds */
  os_boolean bSignalled, bCancelled;

  if (pIoctl == NULL || inSize != sizeof(pIoctl->in) || outSize != sizeof(pIoctl->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate PLX DMA mode */
  mode = pIoctl->in.mode;
  if (mode & 0xffff2600U) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate direction */
  dir = pIoctl->in.direction;
  if (dir != ADMXRC2_PCITOLOCAL && dir != ADMXRC2_LOCALTOPCI) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate local bus address & length */
  pBuffer = pIoctl->in.buffer;
  local = pIoctl->in.local;
  length = pIoctl->in.length;
  if (length == 0) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate the 2D parameters */
  {
    unsigned long n_full_strip, n_skip;
    unsigned long partial_strip;

    width = pIoctl->in.width;
    skip = pIoctl->in.skip;
    localSkip = pIoctl->in.local_skip;

    n_full_strip = length / width;
    partial_strip = length % width;
    if (partial_strip) {
      n_skip = n_full_strip;
    } else {
      n_skip = n_full_strip - 1;
    }
    spanned = width * n_full_strip + partial_strip;
    local_spanned = spanned + n_skip * localSkip;
    spanned += n_skip * skip;
  }

  /* Validate flags parameter */
  flags = pIoctl->in.flags;
  if (flags & ~ADMXRC2_DMAFLAG_DONOTQUEUE) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  /* Validate channel parameter */
  channel = pIoctl->in.channel;
  if (channel != ADMXRC2_DMACHAN_ANY && channel >= devCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  pDmaNode = (DmaNode*)os_alloc_lcache(devCtx->dma.ctlPool, 0);
  if (pDmaNode == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
    goto cleanup;
  }

  /* Get values indicating whether or not to use a timeout */
  bUseTimeout = pIoctl->in.bUseTimeout;
  if (bUseTimeout) {
    start = os_get_time();
    expire = start + os_us_to_ticks(pIoctl->in.timeout);
  } else {
    start = expire = 0; /* Stop compiler complaining about uninit vars. */
  }

  /* Lock down the user buffer and obtain an MDL for it */
  status = ADMXRC_CreateDMATarget(devCtx, pBuffer, spanned, &pDmaTgt);
  if (!OS_SUCCESS(status)) {
    goto cleanup;
  }

  /* Determine which dmaQueue to add this node to */
  queue = (channel == ADMXRC2_DMACHAN_ANY) ? devCtx->info.cardInfo.numDmaChan : (unsigned char) channel;

  /* Prepare the node for insertion into DMA queue */
  pDmaNode->client.pContext = (void*) &event;
  pDmaNode->client.pCallback = doDMACallback;
  pDmaNode->client.pBuffer = pDmaTgt;
  pDmaNode->client.mode = mode;
  pDmaNode->client.local = local;
  pDmaNode->client.length = length;
  pDmaNode->client.offset = 0;
  pDmaNode->client.variant.twoD.width = width;
  pDmaNode->client.variant.twoD.skip = skip;
  pDmaNode->client.variant.twoD.localSkip = localSkip;
  pDmaNode->client.direction = (unsigned char) dir;
  pDmaNode->client.queue = queue;
  pDmaNode->client.bTwoD = TRUE;

  os_dbgprint(6, ("ADMXRC_DoDMAImmediate2D: pDmaNode=0x%p pBuffer=0x%p pDmaTgt=0x%p length=0x%lu local=0x%lx dir=%lu channel=%lu mode=0x%lx flags=0x%lx queue=%lu\n",
		  (void*)pDmaNode,
		  (void*)pBuffer,
		  (void*)pDmaTgt,
		  (unsigned long)length,
		  (unsigned long)local,
		  (unsigned long)dir,
		  (unsigned long)channel,
		  (unsigned long)mode,
		  (unsigned long)flags,
		  (unsigned long)queue));

  status = OS_STATUS_SUCCESS;
  pReq->user.context = (void*) pDmaNode;
  os_init_event(&event);

  bMustStart = (flags & ADMXRC2_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(devCtx, pDmaNode, bMustStart)) {
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
    goto completed;
  }

  if (bUseTimeout) {
    now = os_get_time();
    if (os_time_after(now, expire)) {
      ticks = 0;
    } else {
      ticks = expire - now;
    }
    bCancelled = os_wait_event_timeout(&event, ticks, TRUE, &bSignalled);
    if (bCancelled) {
      if (bSignalled) {
	status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      } else {
	status = OS_MAKE_STATUS(ADMXRC2_TIMEOUT);
      }
      if (!ADMXRC_CancelDma(devCtx, pDmaNode)) {
	os_wait_event(&event, FALSE);
      }
      goto completed;
    }
  } else {
    bSignalled = os_wait_event(&event, TRUE);
    if (bSignalled) {
      status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      if (!ADMXRC_CancelDma(devCtx, pDmaNode)) {
	os_wait_event(&event, FALSE);
      }
      goto completed;
    }
  }

 completed:
  if (OS_SUCCESS(status)) {
    pReq->status.length = sizeof(pIoctl->out);
    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after(now, expire)) {
	pIoctl->out.remaining = 0;
      } else {
	remaining_us = os_ticks_to_us(expire - now);
	if (remaining_us > pIoctl->in.timeout) {
	  pIoctl->out.remaining = pIoctl->in.timeout;
	} else {
	  pIoctl->out.remaining = remaining_us;
	}
      }
    }
  }

 cleanup:
  if (pDmaTgt != NULL) {
    ADMXRC_ReleaseDMATarget(devCtx, pDmaTgt);
  }
  if (pDmaNode != NULL) {
    os_free_lcache(devCtx->dma.ctlPool, pDmaNode);
  }

  return status;
}
