/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2009 Alpha Data 
**
** dmacore2.c - Core functions to support DMA transfers.
**
** ----------------------------------------------------------------------------
**
** Exported functions are:
**
**   ADMXRC_InitDma                Initializes a DMA node
**
**   ADMXRC_EnqueueDma             Queues or begins a DMA transfer
**
**   ADMXRC_CancelDma              Cancels a DMA transfer
**
**   ADMXRC_CleanupDma             Called when shutting down the device; aborts all DMA
**                                 transfers and completes them with a failure status.
**
** Primary datatype is the DmaNode, which is a control structure
** for a DMA operation.
**
** A DMA node's state is as follows:
**
**   QUEUED      => in a DMA queue
**   ACTIVE      => DMA transfer in progress
**   COMPLETED   => completed
**   other       => not a valid state
**
** Substates of ACTIVE regarding cancellation:
**
**   CANCEL_IDLE    => cancellation is not pending
**   CANCEL_PENDING => cancellation is pending
**
** Substates of ACTIVE regarding DMA engine:
**
**   ENGINE_IDLE    => nothing being transferred on hardware
**   ENGINE_RUNNING => data chunk being transferred on hardware
**
** Substates of ACTIVE regarding mapping:
**
**   MAP_IDLE       => nothing being mapped
**   MAP_RUNNING    => a chunk is being mapped
**
** Substates of ACTIVE regarding mapping:
**
**   UNMAP_IDLE     => nothing being unmapped
**   UNMAP_RUNNING  => a chunk is being unmapped
**
** Each DMA node has an attention callback function associated with it,
** called when the DMA transfer transitions to COMPLETED. This callback
** function may be invoked by the DMA core in an arbitrary context.
*/

#include "platform.h"
#include "device.h"
#include "dmacore.h"
#include "dmadesc.h"
#include "drvparam.h"
#include <admxrc2/status.h>

#define DMASTATE_INVALID  (0x0)
#define DMASTATE_QUEUED   (0x1)
#define DMASTATE_ACTIVE   (0x2)
#define DMASTATE_COMPLETE (0x3)

#define DMASTATE_MAP_IDLE    (0x11)
#define DMASTATE_MAP_RUNNING (0x12)

#define DMASTATE_UNMAP_IDLE    (0x21)
#define DMASTATE_UNMAP_RUNNING (0x22)

#define DMASTATE_ENGINE_IDLE    (0x31)
#define DMASTATE_ENGINE_RUNNING (0x32)

#define DMASTATE_CANCEL_IDLE    (0x41)
#define DMASTATE_CANCEL_PENDING (0x42)


/*
** Removes a DMANODE from (anywhere in) a queue.
**
** Must be called while holding the DMA lock.
*/
static void
removeDMA(
  DmaQueue* pQ,
  DmaNode* pNode)
{
  DmaNode* pNext;
  DmaNode* pPrev;

  pNext = pNode->system.pNext;
  pPrev = pNode->system.pPrev;
  if (pNext != NULL) {
    pNext->system.pPrev = pPrev;
  } else {
    pQ->pTail = pPrev;
  }
  if (pPrev != NULL) {
    pPrev->system.pNext = pNext;
  } else {
    pQ->pHead = pNext;
  }
  pQ->length--;
}

/*
** Adds a DMANODE to the tail of a queue.
**
** Must be called while holding the DMA lock.
*/
static void
addDMA(
  DmaQueue* pQ,
  DmaNode* pNode)
{
  DmaNode* pHead;
  DmaNode* pTail;
  
  pHead = pQ->pHead;
  pTail = pQ->pTail;
  if (pHead == NULL) {
    pQ->pTail = pQ->pHead = pNode;
    pNode->system.pNext = NULL;
    pNode->system.pPrev = NULL;
  } else {
    pTail->system.pNext = pNode;
    pQ->pTail = pNode;
    pNode->system.pPrev = pTail;
    pNode->system.pNext = NULL;
  }
  pQ->length++;
}

/*
** Critical section that aborts a DMA transfer on a hardware DMA
** channel. It must not run concurrently with the interrupt handler.
*/
static void*
abortDMAOnHardwareCritical(
  void* arg)
{
  DmaAbortContext* context = (DmaAbortContext*) arg;
  DeviceContext*   pDevCtx = context->devCtx;
  unsigned int     chan = context->channel;
  MethodResult     res;

  res = pDevCtx->methods.doDma(pDevCtx, chan, FALSE, 0, 0);

  context->success = (res == MethodResult_Success);

  return NULL;
}

/*
** Called to abort DMA transfer that is active on the hardware.
**
** Returns TRUE if DMA transfer was successfully aborted, otherwise FALSE.
**
** Must be called while holding the DMA lock.
*/
static os_boolean
abortDMAOnHardware(
  DeviceContext* pDevCtx,
  unsigned char chan)
{
  DmaAbortContext context;

  context.devCtx = pDevCtx;
  context.channel = chan;
  os_invoke_atomic(pDevCtx->irqObj, abortDMAOnHardwareCritical, &context);

  return context.success;
}

/*
** Get & dequeue the next DMANODE, if it exists, on an idle DMA channel.
**
** Looks first for a DMANODE in the queue associated with the
** specified DMA channel, and if not found, looks in the 'any channel'
** queue. If a DMANODE is found, it is made active on the hardware.
**
** Must be called while holding the DMA lock.
*/
static DmaNode*
getNextDMA(
  DeviceContext* pDevCtx,
  unsigned char channel)
{
  DmaNode* pNode;
  DmaControl* pControl;

  os_dbgprint(6, ("getNextDMA: channel=%lu\n", (unsigned long)channel));
  
  /*
  ** The DMA channel is now idle - look for a candidate DMANODE
  ** to make active on the DMA channel.
  */
  pControl = &pDevCtx->dma.channel[channel];
  pNode = pDevCtx->dma.queue[channel].pHead;
  if (pNode == NULL) {
    pNode = pDevCtx->dma.queue[pDevCtx->info.cardInfo.numDmaChan].pHead;
    if (pNode != NULL) {
      /* Found one in the 'any' queue */
      goto found;
    }
  } else {
    /* Found one */
    goto found;
  }

  /* No DMANODE found - can't start a new DMA operation on the hardware */
  return NULL;

 found:
  os_dbgprint(6, ("getNextDMA: pNode=0x%p channel=%lu\n", (void*)pNode, (unsigned long)channel));
  return pNode;
}

/*
** Must be called while NOT holding the DMA lock.
*/
static void
mapChunk(
  DeviceContext* pDevCtx,
  DmaNode* pNode)
{
  DmaControl* pControl;
  DmaMapping* pMapping;
  unsigned char channel;
  unsigned char which;
  size_t offset, chunk, rowChunk, remaining, rowRemaining, position, width, skip, localSkip, length;
  unsigned int descRemaining, index;
  os_uint32 local, mode, mask;
  os_boolean bDmaWrite, bTwoD, bSuccess, bFixedLocal;
  os_uint64 logical;
  void* pKernel;

  os_dbgprint(6, ("mapChunk: pDevCtx=0x%p pNode=0x%p\n", pDevCtx, pNode));

  offset = pNode->client.offset;
  channel = pNode->system.channel;
  pControl = &pDevCtx->dma.channel[channel];
  which = pControl->curr.map.which;
  bDmaWrite = (pNode->client.direction == ADMXRC2_LOCALTOPCI) ? TRUE : FALSE;
  pMapping = pControl->curr.pMapping[which];
  bTwoD = pNode->client.bTwoD;

  chunk = 0;
  remaining = pControl->curr.map.remaining;
  position = pControl->curr.map.position;
  descRemaining = admxrcDriverParams.maxPlxDesc;
  index = 0;
  local = pNode->client.local;
  mode = pNode->client.mode;
  bFixedLocal = (mode & ADMXRC2_DMAMODE_FIXEDLOCAL) ? TRUE : FALSE;
  if (bFixedLocal) {
    mask = (0x1U << (mode & 0x3U)) - 0x1U;
    local = (local & ~mask) | ((local + position) & mask);
  } else {
    mask = 0; /* prevent compiler whine */
    local += position;
  }

  if (bTwoD) {
    /* Map a certain number of pages of the node's two-dimensional buffer */
    width = pNode->client.variant.twoD.width;
    rowRemaining = pControl->curr.map.variant.twoD.rowRemaining;
    skip = pNode->client.variant.twoD.skip;
    localSkip = pNode->client.variant.twoD.localSkip;
    while (index < descRemaining && remaining) {
      rowChunk = (rowRemaining > remaining) ? remaining : rowRemaining;
      if (bFixedLocal) {
	length = local & mask;
	if (length) {
	  /* Workaround for PCI9xxx unaligned address problem with fixed local bus address mode */
	  length = mask - length + 1U;
	  if (length < rowChunk) {
	    rowChunk = length;
	  }
	}
      }
      bSuccess = os_usbd_map(pNode->client.pBuffer->variant.local.pUsbd, pDevCtx->pDevObj, bDmaWrite, position + offset, rowChunk, &pKernel, &logical, &length);
      if (!bSuccess) {
	break;
      }
      
      pMapping->pKernel = pKernel;
      pMapping->logical = logical;
      pMapping->local = local;
      pMapping->length = length;
      pDevCtx->methods.addDmaBuffer(pDevCtx, channel, index, !bDmaWrite, pMapping);
      
      chunk += length;
      remaining -= length;
      position += length;
      if (bFixedLocal) {
	local = (local & ~mask) | ((local + length) & mask);
      } else {
	local += length;
      }
      rowRemaining -= length;
      if (rowRemaining == 0) {
	rowRemaining = width;
	if (mode & ADMXRC2_DMAMODE_FIXEDLOCAL) {
	  local = (local & ~mask) | ((local + localSkip) & mask);
	} else {
	  local += localSkip;
	}
	position += skip;
      }
      index++;
      pMapping++;
    }
    pControl->curr.map.variant.twoD.rowRemaining = rowRemaining;
  } else {
    /* Map a certain number of pages of the node's buffer */
    while (index < descRemaining && remaining) {
      rowChunk = remaining;
      if (bFixedLocal) {
	length = local & mask;
	if (length) {
	  /* Workaround for PCI9xxx unaligned address problem with fixed local bus address mode */
	  length = mask - length + 1U;
	  if (length < rowChunk) {
	    rowChunk = length;
	  }
	}
      }
      bSuccess = os_usbd_map(pNode->client.pBuffer->variant.local.pUsbd, pDevCtx->pDevObj, bDmaWrite, position + offset, rowChunk, &pKernel, &logical, &length);
      if (!bSuccess) {
	break;
      }
      
      pMapping->pKernel = pKernel;
      pMapping->logical = logical;
      pMapping->local = local;
      pMapping->length = length;
      pDevCtx->methods.addDmaBuffer(pDevCtx, channel, index, !bDmaWrite, pMapping);
      
      chunk += length;
      remaining -= length;
      position += length;
      if (bFixedLocal) {
	local = (local & ~mask) | ((local + length) & mask);
      } else {
	local += length;
      }
      index++;
      pMapping++;
    }
  }
  
  pControl->curr.chunk[which] = chunk;
  pControl->curr.nMapped[which] = index;
  pControl->curr.map.position = position;
  pControl->curr.map.remaining = remaining;
  pControl->curr.map.which = which ^ 0x1U;
}
  
/*
** Must be called while holding the DMA lock.
*/
static void
transferChunk(
  DeviceContext* pDevCtx,
  DmaNode* pNode)
{
  DmaControl* pControl;
  unsigned char channel;
  unsigned char which;
  size_t chunk, remaining;
  unsigned int nMapped;

  os_dbgprint(6, ("transferChunk: pDevCtx=0x%p pNode=0x%p\n", pDevCtx, pNode));

  channel = pNode->system.channel;
  pControl = &pDevCtx->dma.channel[channel];
  which = pControl->curr.engine.which;
  nMapped = pControl->curr.nMapped[which];
  remaining = pControl->curr.engine.remaining;
  chunk = pControl->curr.chunk[which];

  /* Start the DMA transfer on the hardware */
  pDevCtx->methods.doDma(pDevCtx, channel, TRUE, pNode->client.mode, nMapped);

  remaining -= chunk;
  pControl->curr.engine.remaining = remaining;
  pControl->curr.engine.which = which ^ 0x1U;
}
  
/*
** Must be called while NOT holding the DMA lock.
*/
static void
unmapChunk(
  DeviceContext* pDevCtx,
  DmaNode* pNode)
{
  DmaControl* pControl;
  DmaMapping* pMapping;
  unsigned char channel;
  unsigned char which;
  unsigned int nMapped;
  size_t chunk, remaining;
  os_boolean bDmaWrite;

  os_dbgprint(6, ("unmapChunk: pDevCtx=0x%p pNode=0x%p\n", pDevCtx, pNode));

  channel = pNode->system.channel;
  pControl = &pDevCtx->dma.channel[channel];
  which = pControl->curr.unmap.which;
  bDmaWrite = (pNode->client.direction == ADMXRC2_LOCALTOPCI) ? TRUE : FALSE;
  
  /* Undo a mapping of a certain number of pages of the node's buffer */
  chunk = pControl->curr.chunk[which];
  nMapped = pControl->curr.nMapped[which];
  pMapping = pControl->curr.pMapping[which];
  remaining = pControl->curr.unmap.remaining;

  while (nMapped) {
    uint32_t length;
    uint64_t logical;
    void* pKernel;

    pKernel = pMapping->pKernel;
    logical = pMapping->logical;
    length = pMapping->length;
    os_usbd_unmap(pNode->client.pBuffer->variant.local.pUsbd, pDevCtx->pDevObj, bDmaWrite, pKernel, logical, length);

    nMapped--;
    pMapping++;
  }

  remaining -= chunk;
  pControl->curr.unmap.remaining = remaining;
  pControl->curr.unmap.which = which ^ 0x1U;
}
  
/*
** Must be called while holding the DMA lock.
*/
static os_boolean
bCanFinishCancel(
  DmaControl* pControl)
{
  return (pControl->curr.map.state == DMASTATE_MAP_IDLE &&
	  pControl->curr.unmap.state == DMASTATE_UNMAP_IDLE &&
	  pControl->curr.engine.state == DMASTATE_ENGINE_IDLE) ? TRUE : FALSE;
}

/*
** Must be called while holding the DMA lock.
*/
static os_boolean
bCanDoChunk(
  DmaControl* pControl)
{
  return (pControl->curr.engine.state == DMASTATE_ENGINE_IDLE &&
	  pControl->curr.engine.remaining > pControl->curr.map.remaining) ? TRUE : FALSE;
}

/*
** Must be called while holding the DMA lock.
*/
static os_boolean
bCanMapChunk(
  DmaControl* pControl)
{
  return (pControl->curr.map.remaining > 0 &&
	  pControl->curr.map.state == DMASTATE_MAP_IDLE &&
	  pControl->curr.mapped < 2) ? TRUE : FALSE;
}

#if 0
/*
** Must be called while holding the DMA lock.
*/
static os_boolean
bCanUnmapChunk(
  DmaControl* pControl)
{
  return (pControl->curr.unmap.remaining > 0 &&
	  pControl->curr.unmap.state == DMASTATE_UNMAP_IDLE &&
	  pControl->curr.mapped > 0) ? TRUE : FALSE;
}
#endif

/*
** Must be called while holding the DMA lock.
**
** 'pNode' must be a DMA node in the ACTIVE state.
**
** Completes cancellation of a DMA node and returns the next DMA node to become
** ACTIVE, if there is such a DMA node.
*/
static DmaNode*
finishCancel(
  DeviceContext* pDevCtx,
  DmaNode* pNode)
{
  unsigned char channel;

  os_dbgprint(6, ("finishCancel: pDevCtx=0x%p pNode=0x%p", pDevCtx, pNode));

  /* Everything else is idle, so we can finish cancellation */
  channel = pNode->system.channel;
  pNode->system.state = DMASTATE_COMPLETE;
  pNode->status.reason = DMAREASON_CANCELLED;
  pNode->status.status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
  
  return getNextDMA(pDevCtx, channel);
}

/*
** Called when a DMA node transitions to the ACTIVE state.
**
** On entry:
** - 'pNode->mapped' must be 0
** - 'pNode->state' must be ACTIVE
** - 'pNode->map.state' must be MAP_RUNNING
** - 'pNode->unmap.state' must be UNMAP_IDLE
** - 'pNode->engine.state' must be ENGINE_IDLE
** - DMA spinlock must not be held.
*/
static void
beginTransfer(
  DeviceContext* pDevCtx,
  DmaNode* pNode)
{
  DmaNode* pNextNode;
  DmaControl* pControl;
  OS_SPINLOCK_FLAGS f;
  unsigned char channel, queue;
  unsigned int unmap;

  os_dbgprint(6, ("beginTransfer: pDevCtx=0x%p pNode=0x%p", pDevCtx, pNode));

  channel = pNode->system.channel;
  pControl = &pDevCtx->dma.channel[channel];

  while (1) {
    mapChunk(pDevCtx, pNode);
  
    os_acquire_spinlock(&pDevCtx->dma.lock, &f);
    pControl->curr.mapped++;
    pControl->curr.map.state = DMASTATE_MAP_IDLE;
    if (pControl->curr.cancel.state == DMASTATE_CANCEL_PENDING) {
      if (bCanFinishCancel(pControl)) {
	/* Everything else is idle, so we can finish cancellation */
	unmap = pControl->curr.mapped;
	pControl->curr.mapped = 0;
	pNextNode = finishCancel(pDevCtx, pNode);
	pControl->curr.pNode = pNextNode;
	if (NULL != pNextNode) {
	  queue = pNextNode->client.queue;
	  removeDMA(&pDevCtx->dma.queue[queue], pNextNode);
	  pNextNode->system.channel = channel;
	  pNextNode->system.state = DMASTATE_ACTIVE;
	  pControl->curr.cancel.state = DMASTATE_CANCEL_IDLE;
	  pControl->curr.map.state = DMASTATE_MAP_RUNNING;
	  pControl->curr.map.position = 0;
	  pControl->curr.map.remaining = pNextNode->client.length;
	  pControl->curr.map.variant.twoD.rowRemaining = pNextNode->client.variant.twoD.width;
	  pControl->curr.map.which = 0;
	  pControl->curr.engine.state = DMASTATE_ENGINE_IDLE;
	  pControl->curr.engine.remaining = pNextNode->client.length;
	  pControl->curr.engine.which = 0;
	}

	os_release_spinlock(&pDevCtx->dma.lock, f); 

	while (unmap--) {
	  unmapChunk(pDevCtx, pNode);
	}

	pNode->client.pCallback(pDevCtx, pNode);
	
	pNode = pNextNode;
      } else {
	/* Not everything else is idle */
	os_release_spinlock(&pDevCtx->dma.lock, f); 
	break;
      }
    } else {
      if (bCanDoChunk(pControl)) {
	pControl->curr.engine.state = DMASTATE_ENGINE_RUNNING;
	transferChunk(pDevCtx, pNode);
      }
      if (bCanMapChunk(pControl)) {
	/* This DMA transfer involves multiple chunks */
	pControl->curr.map.state = DMASTATE_MAP_RUNNING;
	os_release_spinlock(&pDevCtx->dma.lock, f);
      } else {
	/* Nothing more to do */
	os_release_spinlock(&pDevCtx->dma.lock, f);
	break;
      }
    }
  }
}

/*
** Initialize a DMA node
*/
void
ADMXRC_InitDma(
  DmaNode* pNode)
{
  pNode->system.state = DMASTATE_INVALID;
}

/*
** Queue up a DMA node
*/
os_boolean
ADMXRC_EnqueueDma(
  DeviceContext* pDevCtx,
  DmaNode* pNode,
  os_boolean bMustStart)
{
  OS_SPINLOCK_FLAGS f;
  unsigned char queue;
  unsigned char channel;
  os_boolean bError = FALSE;
  os_boolean bCanStart = FALSE;
  DmaControl* pControl = NULL;

  queue = pNode->client.queue;
  os_dbgprint(6, ("ADMXRC_EnqueueDMA: pNode=0x%p queue=%lu\n",
		  (void*)pNode, (unsigned long)queue));

  os_acquire_spinlock(&pDevCtx->dma.lock, &f);

  /*
  ** Attempt to bypass queueing mechanism - check if there is a DMA channel
  ** free on which to start the DMA transfer immediately
  */
  if (queue < pDevCtx->info.cardInfo.numDmaChan) {
    /* Must use a specific DMA channel */
    channel = queue;
    pControl = &pDevCtx->dma.channel[channel];
    if (pControl->curr.pNode == NULL && pDevCtx->dma.queue[channel].pHead == NULL) {
      bCanStart = TRUE;
    }
  } else {
    /* Can use any DMA channel */
    for (channel = 0; channel < pDevCtx->info.cardInfo.numDmaChan; channel++) {
      pControl = &pDevCtx->dma.channel[channel];
      if (pControl->curr.pNode == NULL && pDevCtx->dma.queue[channel].pHead == NULL) {
	bCanStart = TRUE;
	break;
      }
    }
  }

  if (bCanStart) {
    pNode->system.channel = channel;
    pNode->system.state = DMASTATE_ACTIVE;
    pControl->curr.cancel.state = DMASTATE_CANCEL_IDLE;
    pControl->curr.map.state = DMASTATE_MAP_RUNNING;
    pControl->curr.map.position = 0;
    pControl->curr.map.remaining = pNode->client.length;
    pControl->curr.map.variant.twoD.rowRemaining = pNode->client.variant.twoD.width;
    pControl->curr.map.which = 0;
    pControl->curr.unmap.state = DMASTATE_UNMAP_IDLE;
    pControl->curr.unmap.remaining = pNode->client.length;
    pControl->curr.unmap.which = 0;
    pControl->curr.engine.state = DMASTATE_ENGINE_IDLE;
    pControl->curr.engine.remaining = pNode->client.length;
    pControl->curr.engine.which = 0;
    pControl->curr.pNode = pNode;
    os_release_spinlock(&pDevCtx->dma.lock, f);

    beginTransfer(pDevCtx, pNode);
  } else {
    if (bMustStart) {
      os_dbgprint(6, ("ADMXRC_EnqueueDMA: cannot start immediately\n"));
      pNode->status.status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
      bError = TRUE;
    } else {
      /*
      ** No free DMA channel on which we can start this DMA operation
      ** immediately, so add the DMANODE to the tail of the DMA queue.
      */
      queue = pNode->client.queue;
      os_dbgprint(6, ("ADMXRC_EnqueueDMA: queueing DMA node, pNode=0x%p queue=%lu\n",
		      (void*)pNode, (unsigned long)queue));
      addDMA(&pDevCtx->dma.queue[queue], pNode);
      pNode->system.state = DMASTATE_QUEUED;      
    }
    os_release_spinlock(&pDevCtx->dma.lock, f);
  }

  return bError;
}

/*
** This function is called to attempt to cancel a DMA transfer.
*/
os_boolean
ADMXRC_CancelDma(
  DeviceContext* pDevCtx,
  DmaNode* pNode)
{
  os_boolean bSuccess;
  OS_SPINLOCK_FLAGS f;
  unsigned char channel;
  unsigned char queue;
  unsigned int unmap = 0;
  int state;
  DmaNode* pNextNode = NULL;
  DmaControl* pControl = NULL;

  os_acquire_spinlock(&pDevCtx->dma.lock, &f);
  state = pNode->system.state;
  channel = pNode->system.channel;
  switch (state) {
  case DMASTATE_QUEUED:
    queue = pNode->client.queue;
    removeDMA(&pDevCtx->dma.queue[queue], pNode);
    pNode->status.reason = DMAREASON_CANCELLED;
    pNode->status.status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
    bSuccess = TRUE;
    os_dbgprint(1, ("ADMXRC_CancelDMA: DMA node in QUEUED state cancelled successfully, pNode=0x%p\n", (void*)pNode));
    break;

  case DMASTATE_ACTIVE:
    pControl = &pDevCtx->dma.channel[channel];
    if (pControl->curr.engine.state == DMASTATE_ENGINE_RUNNING) {
      abortDMAOnHardware(pDevCtx, channel);
      pControl->curr.engine.state = DMASTATE_ENGINE_IDLE;
    }
    if (bCanFinishCancel(pControl)) {
      pNextNode = finishCancel(pDevCtx, pNode);
      os_dbgprint(1, ("ADMXRC_CancelDMA: DMA node in ACTIVE state cancelled successfully, pNode=0x%p pNextNode=0x%p\n",
		      (void*)pNode, (void*)pNextNode));
      bSuccess = TRUE;
      unmap = pControl->curr.mapped;
      pControl->curr.mapped = 0;
      pControl->curr.pNode = pNextNode;
      if (pNextNode != NULL) {
	queue = pNextNode->client.queue;
	removeDMA(&pDevCtx->dma.queue[queue], pNextNode);
	pNextNode->system.channel = channel;
	pNextNode->system.state = DMASTATE_ACTIVE;
	pControl->curr.cancel.state = DMASTATE_CANCEL_IDLE;
	pControl->curr.map.state = DMASTATE_MAP_RUNNING;
	pControl->curr.map.position = 0;
	pControl->curr.map.remaining = pNextNode->client.length;
	pControl->curr.map.variant.twoD.rowRemaining = pNextNode->client.variant.twoD.width;
	pControl->curr.map.which = 0;
	pControl->curr.engine.state = DMASTATE_ENGINE_IDLE;
	pControl->curr.engine.remaining = pNextNode->client.length;
	pControl->curr.engine.which = 0;
      }
    } else {
      os_dbgprint(1, ("ADMXRC_CancelDMA: deferring cancellation of DMA node ACTIVE state, pNode=0x%p\n", (void*)pNode));
      os_dbgprint(1, ("ADMXRC_CancelDMA: engineState=%u mapState=%u unmapState=%u\n",
		      pControl->curr.engine.state,
		      pControl->curr.map.state,
		      pControl->curr.unmap.state));
      pControl->curr.cancel.state = DMASTATE_CANCEL_PENDING;
      bSuccess = FALSE;
    }
    break;

  case DMASTATE_COMPLETE:
    /*
    ** The ball is currently in the client's court. The DMA callback has been
    ** invoked, is being invoked, or is about to be invoked with COMPLETED reason.
    **
    ** So don't call the DMA callback.
    */
    os_dbgprint(1, ("ADMXRC_CancelDMA: cancellation of DMA node 0x%p in COMPLETED state not required\n", (void*)pNode));
    bSuccess = FALSE;
    break;

  default: /* Should never happen! */
    os_dbgprint(0, ("*** ADMXRC_CancelDMA: DMA node 0x%p has invalid state (%lu)\n",
		    (void*)pNode, (unsigned long)state));
    bSuccess = FALSE;
    break;
  }
  os_release_spinlock(&pDevCtx->dma.lock, f);
  
  if (bSuccess) {
    if (NULL != pControl) {
      while (unmap--) {
	unmapChunk(pDevCtx, pNode);
      }

      if (pNextNode != NULL) {
	os_dbgprint(1, ("ADMXRC_CancelDMA: starting next node, pNextNode=0x%p\n", (void*)pNextNode));
	pControl->curr.unmap.state = DMASTATE_UNMAP_IDLE;
	pControl->curr.unmap.remaining = pNextNode->client.length;
	pControl->curr.unmap.which = 0;
	beginTransfer(pDevCtx, pNextNode);
      }
    }
  }

  return bSuccess;
}

/*
** This routine is called by the DMA interrupt DPC to service a DMA interrupt
** for a particular channel.
**
** May be called outside a thread context.
*/
void
ADMXRC_DmaCompleted(
 DeviceContext* pDevCtx,
 unsigned char channel)
{
  OS_SPINLOCK_FLAGS f;
  DmaNode* pNode;
  DmaNode* pNextNode;
  DmaControl* pControl;
  unsigned char queue;
  unsigned int unmap = 0;

  os_acquire_spinlock(&pDevCtx->dma.lock, &f);
  pControl = &pDevCtx->dma.channel[channel];
  pNode = pControl->curr.pNode;
  if (pNode == NULL) {
    os_release_spinlock(&pDevCtx->dma.lock, f);
    
    os_dbgprint(1, ("*** ADMXRC_DMACompleted: no job on channel %lu!\n", (unsigned long)channel));
  } else {
    pControl->curr.engine.state = DMASTATE_ENGINE_IDLE;
    pControl->curr.unmap.state = DMASTATE_UNMAP_RUNNING;
    if (bCanDoChunk(pControl)) {
      pControl->curr.engine.state = DMASTATE_ENGINE_RUNNING;
      transferChunk(pDevCtx, pNode);
    }
    os_release_spinlock(&pDevCtx->dma.lock, f);
    
    unmapChunk(pDevCtx, pNode);

    os_acquire_spinlock(&pDevCtx->dma.lock, &f);
    pControl->curr.unmap.state = DMASTATE_UNMAP_IDLE;
    pControl->curr.mapped--;
    if (pControl->curr.cancel.state == DMASTATE_CANCEL_PENDING) {
      if (bCanFinishCancel(pControl)) {
	os_dbgprint(1, ("ADMXRC_DMACompleted: finishing cancellation of DMA node in ACTIVE state, pNode=0x%p\n", (void*)pNode));
	pNextNode = finishCancel(pDevCtx, pNode);
	unmap = pControl->curr.mapped;
	pControl->curr.mapped = 0;
	pControl->curr.pNode = pNextNode;
	if (NULL != pNextNode) {
	  queue = pNextNode->client.queue;
	  removeDMA(&pDevCtx->dma.queue[queue], pNextNode);
	  pNextNode->system.channel = channel;
	  pNextNode->system.state = DMASTATE_ACTIVE;
	  pControl->curr.cancel.state = DMASTATE_CANCEL_IDLE;
	  pControl->curr.map.state = DMASTATE_MAP_RUNNING;
	  pControl->curr.map.position = 0;
	  pControl->curr.map.remaining = pNextNode->client.length;
	  pControl->curr.map.variant.twoD.rowRemaining = pNextNode->client.variant.twoD.width;
	  pControl->curr.map.which = 0;
	  pControl->curr.engine.state = DMASTATE_ENGINE_IDLE;
	  pControl->curr.engine.remaining = pNextNode->client.length;
	  pControl->curr.engine.which = 0;
	}
	os_release_spinlock(&pDevCtx->dma.lock, f); 
     
	while (unmap--) {
	  unmapChunk(pDevCtx, pNode);
	}

	pNode->client.pCallback(pDevCtx, pNode);
	
	if (NULL != pNextNode) {
	  pControl->curr.unmap.state = DMASTATE_UNMAP_IDLE;
	  pControl->curr.unmap.remaining = pNextNode->client.length;
	  pControl->curr.unmap.which = 0;
	  beginTransfer(pDevCtx, pNextNode);
	}
      } else {
	/* Not everything else is idle */
	os_dbgprint(1, ("ADMXRC_DMACompleted: deferring cancellation of DMA node in ACTIVE state, pNode=0x%p\n", (void*)pNode));
	os_release_spinlock(&pDevCtx->dma.lock, f); 
      }
    } else {
      if (bCanMapChunk(pControl)) {
	pControl->curr.map.state = DMASTATE_MAP_RUNNING;
	os_release_spinlock(&pDevCtx->dma.lock, f);

	mapChunk(pDevCtx, pNode);

	os_acquire_spinlock(&pDevCtx->dma.lock, &f);
	pControl->curr.map.state = DMASTATE_MAP_IDLE;
	pControl->curr.mapped++;
	if (pControl->curr.cancel.state == DMASTATE_CANCEL_PENDING) {
	  if (bCanFinishCancel(pControl)) {
	    os_dbgprint(1, ("ADMXRC_DMACompleted: finishing cancellation of DMA node in ACTIVE state, pNode=0x%p\n", (void*)pNode));
	    pNextNode = finishCancel(pDevCtx, pNode);
	    unmap = pControl->curr.mapped;
	    pControl->curr.mapped = 0;
	    pControl->curr.pNode = pNextNode;
	    if (NULL != pNextNode) {
	      queue = pNextNode->client.queue;
	      removeDMA(&pDevCtx->dma.queue[queue], pNextNode);
	      pNextNode->system.channel = channel;
	      pNextNode->system.state = DMASTATE_ACTIVE;
	      pControl->curr.cancel.state = DMASTATE_CANCEL_IDLE;
	      pControl->curr.map.state = DMASTATE_MAP_RUNNING;
	      pControl->curr.map.position = 0;
	      pControl->curr.map.remaining = pNextNode->client.length;
	      pControl->curr.map.variant.twoD.rowRemaining = pNextNode->client.variant.twoD.width;
	      pControl->curr.map.which = 0;
	      pControl->curr.engine.state = DMASTATE_ENGINE_IDLE;
	      pControl->curr.engine.remaining = pNextNode->client.length;
	      pControl->curr.engine.which = 0;
	    }
	    os_release_spinlock(&pDevCtx->dma.lock, f); 
	    
	    while (unmap--) {
	      unmapChunk(pDevCtx, pNode);
	    }

	    pNode->client.pCallback(pDevCtx, pNode);
	    
	    if (NULL != pNextNode) {
	      pControl->curr.unmap.state = DMASTATE_UNMAP_IDLE;
	      pControl->curr.unmap.remaining = pNextNode->client.length;
	      pControl->curr.unmap.which = 0;
	      beginTransfer(pDevCtx, pNextNode);
	    }
	  } else {
	    /* Not everything else is idle */
	    os_dbgprint(1, ("ADMXRC_DMACompleted: deferring cancellation of DMA node in ACTIVE state, pNode=0x%p\n", (void*)pNode));
	    os_release_spinlock(&pDevCtx->dma.lock, f); 
	  }
	} else {
	  if (bCanDoChunk(pControl)) {
	    pControl->curr.engine.state = DMASTATE_ENGINE_RUNNING;
	    transferChunk(pDevCtx, pNode);
	    os_release_spinlock(&pDevCtx->dma.lock, f);
	  } else {
	    /* Nothing to do */
	    os_release_spinlock(&pDevCtx->dma.lock, f);
	  }
	}
      } else {
	if (pControl->curr.unmap.remaining == 0) {
	  /* The entire DMA transfer has been completed */
	  pNode->system.state = DMASTATE_COMPLETE;
	  pNode->status.reason = DMAREASON_COMPLETE;
	  pNode->status.status = OS_STATUS_SUCCESS;
	  pNextNode = getNextDMA(pDevCtx, channel);
	  pControl->curr.pNode = pNextNode;
	  if (NULL != pNextNode) {
	    queue = pNextNode->client.queue;
	    removeDMA(&pDevCtx->dma.queue[queue], pNextNode);
	    pNextNode->system.channel = channel;
	    pNextNode->system.state = DMASTATE_ACTIVE;
	    pControl->curr.cancel.state = DMASTATE_CANCEL_IDLE;
	    pControl->curr.map.state = DMASTATE_MAP_RUNNING;
	    pControl->curr.map.position = 0;
	    pControl->curr.map.remaining = pNextNode->client.length;
	    pControl->curr.map.variant.twoD.rowRemaining = pNextNode->client.variant.twoD.width;
	    pControl->curr.map.which = 0;
	    pControl->curr.unmap.state = DMASTATE_UNMAP_IDLE;
	    pControl->curr.unmap.remaining = pNextNode->client.length;
	    pControl->curr.unmap.which = 0;
	    pControl->curr.engine.state = DMASTATE_ENGINE_IDLE;
	    pControl->curr.engine.remaining = pNextNode->client.length;
	    pControl->curr.engine.which = 0;
	  }
	  os_release_spinlock(&pDevCtx->dma.lock, f);
	  
	  pNode->client.pCallback(pDevCtx, pNode);

	  if (NULL != pNextNode) {
	    os_dbgprint(6, ("ADMXRC_DMACompleted: starting next node, pNextNode=0x%p\n", (void*)pNextNode));
	    beginTransfer(pDevCtx, pNextNode);
	  }
	} else {
	  /* Nothing to do */
	  os_release_spinlock(&pDevCtx->dma.lock, f);
	}
      }
    }
  }
}

/*
** DMA cleanup function; called when card is about to be closed
*/
void
ADMXRC_CleanupDma(
  DeviceContext* pDevCtx)
{
  OS_SPINLOCK_FLAGS f;
  DmaNode* pNode;
  DmaControl* pControl;
  unsigned char i;
  unsigned int unmap;

  if (os_dbglevel(1)) {
    for (i = 0; i <= pDevCtx->info.cardInfo.numDmaChan; i++) {
      os_dbgprint(1, ("ADMXRC_CleanupDMA: queue %lu length=%lu\n",
		      (unsigned long)i, (unsigned long)pDevCtx->dma.queue[i].length));
    }
  }

  /* Cleanup any DMA transfers that are in the QUEUED state */
  for (i = 0; i <= pDevCtx->info.cardInfo.numDmaChan; i++) {
    while (1) {
      os_acquire_spinlock(&pDevCtx->dma.lock, &f);
      pNode = pDevCtx->dma.queue[i].pHead;
      if (pNode == NULL) {
	os_release_spinlock(&pDevCtx->dma.lock, f);
	break;
      } else {
	os_dbgprint(1, ("ADMXRC_CleanupDMA: Cleaning up DMANODE 0x%p on DMA queue %lu\n",
			(void*)pNode, (unsigned long)i));
	
	/* Remove this DMANODE from the queue */
	removeDMA(&pDevCtx->dma.queue[i], pNode);
	
	/* Mark the DMANODE as finished, but with an error status */
	pNode->system.state = DMASTATE_COMPLETE;
	pNode->status.reason = DMAREASON_CANCELLED;
	pNode->status.status = OS_MAKE_STATUS(ADMXRC2_CLOSED);
	os_release_spinlock(&pDevCtx->dma.lock, f);
	
	pNode->client.pCallback(pDevCtx, pNode);
      }
    }
  }

  /* Cleanup any DMA transfers that are currently in the ACTIVE state */
  for (i = 0; i < pDevCtx->info.cardInfo.numDmaChan; i++) {
    os_acquire_spinlock(&pDevCtx->dma.lock, &f);
    pControl = &pDevCtx->dma.channel[i];
    pNode = pControl->curr.pNode;
    if (pNode != NULL) {
      if (pControl->curr.engine.state == DMASTATE_ENGINE_RUNNING) {
	abortDMAOnHardware(pDevCtx, i);
	pControl->curr.engine.state = DMASTATE_ENGINE_IDLE;
      }
      if (bCanFinishCancel(pControl)) {
	/* We can finish cancelling this node */
	finishCancel(pDevCtx, pNode);
	os_dbgprint(1, ("ADMXRC_CleanupDMA: cleaned up successfully, pNode=0x%p\n", (void*)pNode));
	unmap = pControl->curr.mapped;
	pControl->curr.mapped = 0;
	pControl->curr.pNode = NULL;
	os_release_spinlock(&pDevCtx->dma.lock, f);

	while (unmap--) {
	  unmapChunk(pDevCtx, pNode);
	}

	pNode->client.pCallback(pDevCtx, pNode);
      } else {
	/* Can't finish cancellation at the moment */
	os_dbgprint(1, ("ADMXRC_CleanupDMA: deferring cleanup, pNode=0x%p\n", (void*)pNode));
	pControl->curr.cancel.state = DMASTATE_CANCEL_PENDING;
	os_release_spinlock(&pDevCtx->dma.lock, f);
      }
    } else {
      /* Nothing to do */
      os_release_spinlock(&pDevCtx->dma.lock, f);
    }
  }
  
  os_dbgprint(1, ("ADMXRC_CleanupDMA: done\n"));
}
